@@ -101,7 +101,7 @@ int main(int argc, char *argv[])
 	close(spair[1]);
 
 	// Drop privileges
-	fz::impersonation_token nobody("nobody", fz::impersonation_flag::pwless);
+	fz::impersonation_token nobody("nobody", fz::impersonation_options::pwless);
 	if (!nobody) {
 		std::cerr << "Parent: Could not get token for \"nobody\"\n";
 		abort();
