@@ -45,6 +45,9 @@ int main(int argc, char *argv[])
 		if (!time.empty()) {
 			std::cout << "    Last modified: " << time.format("%Y-%m-%d %H-%M-%S ", fz::datetime::local) << "\n";
 		}
+		if (is_link) {
+			std::cout << "    Target of link: " << fz::to_string(fz::local_filesys::get_link_target(path + fzT("/") + name)) << "\n";
+		}
 		std::cout << "    Mode: " << mode << "\n" << std::endl;
 	}
 
