@@ -10,6 +10,7 @@ void reader_base::close()
 	scoped_lock l(mtx_);
 	do_close(l);
 	buffer_pool_.remove_waiter(*this);
+	waiting_ = false;
 	remove_waiters();
 	buffers_.clear();
 }
@@ -80,6 +81,7 @@ bool reader_base::seek(uint64_t offset, uint64_t size)
 	}
 
 	buffer_pool_.remove_waiter(*this);
+	waiting_ = false;
 	remove_waiters();
 	buffers_.clear();
 
@@ -107,26 +109,31 @@ bool reader_base::error() const
 	return error_;
 }
 
-std::pair<aio_result, buffer_lease> reader_base::get_buffer(aio_waiter & h)
-{
-	scoped_lock l(mtx_);
-	auto ret = do_get_buffer(l);
-	if (ret.first == aio_result::wait) {
-		add_waiter(h);
-	}
-	return ret;
-}
-
 std::pair<aio_result, buffer_lease> reader_base::get_buffer(event_handler & h)
 {
 	scoped_lock l(mtx_);
 	auto ret = do_get_buffer(l);
 	if (ret.first == aio_result::wait) {
+		//fz::sleep(duration::from_milliseconds(1000));
 		add_waiter(h);
 	}
 	return ret;
 }
 
+void reader_base::operator()(event_base const& ev)
+{
+	dispatch<aio_buffer_event>(ev, this, &reader_base::on_buffer_availability);
+}
+
+void reader_base::on_buffer_availability(aio_waitable const* w)
+{
+	fz::scoped_lock l(mtx_);
+	if (waiting_) {
+		waiting_ = false;
+		do_on_buffer_availability(l, w);
+	}
+}
+
 reader_factory_holder::reader_factory_holder(reader_factory_holder const& op)
 {
 	if (op.impl_) {
@@ -251,6 +258,7 @@ file_reader::file_reader(std::wstring_view name, aio_buffer_pool & pool, file &&
 
 file_reader::~file_reader() noexcept
 {
+	remove_handler();
 	close();
 }
 
@@ -294,9 +302,8 @@ bool file_reader::do_seek(scoped_lock & l)
 	}
 }
 
-void file_reader::on_buffer_availability(aio_waitable const*)
+void file_reader::do_on_buffer_availability(fz::scoped_lock& l, aio_waitable const*)
 {
-	scoped_lock l(mtx_);
 	cond_.signal(l);
 }
 
@@ -310,6 +317,7 @@ void file_reader::entry()
 		}
 		auto b = buffer_pool_.get_buffer(*this);
 		if (!b) {
+			waiting_ = true;
 			cond_.wait(l);
 			continue;
 		}
@@ -436,6 +444,7 @@ view_reader::view_reader(std::wstring && name, aio_buffer_pool & pool, std::stri
 
 view_reader::~view_reader() noexcept
 {
+	remove_handler();
 	close();
 }
 
@@ -454,6 +463,7 @@ std::pair<aio_result, buffer_lease> view_reader::do_get_buffer(scoped_lock &)
 
 	auto b = buffer_pool_.get_buffer(*this);
 	if (!b) {
+		waiting_ = true;
 		return {aio_result::wait, buffer_lease()};
 	}
 
@@ -471,7 +481,7 @@ std::pair<aio_result, buffer_lease> view_reader::do_get_buffer(scoped_lock &)
 	return {aio_result::ok, std::move(b)};
 }
 
-void view_reader::on_buffer_availability(aio_waitable const*)
+void view_reader::do_on_buffer_availability(fz::scoped_lock&, aio_waitable const*)
 {
 	signal_availibility();
 }
@@ -523,6 +533,7 @@ string_reader::string_reader(std::wstring && name, aio_buffer_pool & pool, std::
 
 string_reader::~string_reader() noexcept
 {
+	remove_handler();
 	close();
 }
 
@@ -541,6 +552,7 @@ std::pair<aio_result, buffer_lease> string_reader::do_get_buffer(scoped_lock &)
 
 	auto b = buffer_pool_.get_buffer(*this);
 	if (!b) {
+		waiting_ = true;
 		return {aio_result::wait, buffer_lease()};
 	}
 
@@ -558,7 +570,7 @@ std::pair<aio_result, buffer_lease> string_reader::do_get_buffer(scoped_lock &)
 	return {aio_result::ok, std::move(b)};
 }
 
-void string_reader::on_buffer_availability(aio_waitable const*)
+void string_reader::do_on_buffer_availability(fz::scoped_lock&, aio_waitable const*)
 {
 	signal_availibility();
 }
