@@ -17,24 +17,23 @@ namespace fz {
 event_loop::event_loop()
 	: sync_(false)
 	, thread_(std::make_unique<thread>())
-	, timer_thread_(std::make_unique<thread>())
+	, mode_(Mode::thread)
 {
 	thread_->run([this] { entry(); });
-	timer_thread_->run([this] { timer_entry(); });
 }
 
 event_loop::event_loop(thread_pool & pool)
 	: sync_(false)
+	, pool_(&pool)
+	, mode_(Mode::tasks)
 {
 	task_ = std::make_unique<async_task>(pool.spawn([this] { entry(); }));
-	timer_task_ = std::make_unique<async_task>(pool.spawn([this] { timer_entry(); }));
 }
 
 event_loop::event_loop(event_loop::loop_option)
 	: sync_(false)
-	, timer_thread_(std::make_unique<thread>())
+	, mode_(Mode::threadless)
 {
-	timer_thread_->run([this] { timer_entry(); });
 }
 
 event_loop::~event_loop()
@@ -136,19 +135,19 @@ void event_loop::filter_events(std::function<bool(event_handler*& h, event_base&
 timer_id event_loop::add_timer(event_handler* handler, monotonic_clock const &deadline, duration const& interval)
 {
 	timer_id id = 0;
-	
+
 	if (deadline) {
 		timer_data d;
 
 		scoped_lock lock(sync_);
-		
+
 		id = setup_timer(lock, d, handler, deadline, interval);
 
 		if (id) {
 			timers_.push_back(std::move(d));
 		}
 	}
-	
+
 	return id;
 }
 
@@ -175,7 +174,7 @@ void event_loop::stop_timer(timer_id id)
 timer_id event_loop::stop_add_timer(timer_id id, event_handler* handler, monotonic_clock const &deadline, duration const& interval)
 {
 	scoped_lock lock(sync_);
-	
+
 	if (id) {
 		for (auto it = timers_.begin(); it != timers_.end(); ++it) {
 			if (it->id_ == id) {
@@ -195,7 +194,7 @@ timer_id event_loop::stop_add_timer(timer_id id, event_handler* handler, monoton
 	return id;
 }
 
-timer_id event_loop::setup_timer(scoped_lock &l, timer_data &d, event_handler* handler,  monotonic_clock const& deadline, duration const& interval)
+timer_id event_loop::setup_timer(scoped_lock &l, timer_data &d, event_handler* handler, monotonic_clock const& deadline, duration const& interval)
 {
 	if (handler->removing_) {
 		return 0;
@@ -209,7 +208,26 @@ timer_id event_loop::setup_timer(scoped_lock &l, timer_data &d, event_handler* h
 	if (!deadline_ || d.deadline_ < deadline_) {
 		// Our new time is the next timer to trigger
 		deadline_ = d.deadline_;
-		timer_cond_.signal(l);
+
+		switch (mode_) {
+		case Mode::thread:
+			if (!timer_thread_) {
+				timer_thread_ = std::make_unique<thread>();
+				timer_thread_->run([this] { timer_entry(); });
+			}
+			timer_cond_.signal(l);
+			break;
+		case Mode::tasks:
+			if (!timer_task_) {
+				timer_task_ = std::make_unique<async_task>(pool_->spawn([this] { timer_entry(); }));
+			}
+			timer_cond_.signal(l);
+			break;
+		default:
+			do_timers_ = true;
+			cond_.signal(l);
+			break;
+		}
 	}
 
 	return d.id_;
@@ -296,7 +314,12 @@ void event_loop::entry()
 		}
 
 		// Nothing to do, now we wait
-		cond_.wait(l);
+		if (threadless_ && deadline_) {
+			cond_.wait(l, deadline_ - fz::monotonic_clock::now());
+		}
+		else {
+			cond_.wait(l);
+		}
 	}
 }
 
@@ -327,7 +350,9 @@ void event_loop::timer_entry()
 bool event_loop::process_timers(scoped_lock & l)
 {
 	if (!deadline_) {
-		do_timers_ = false;
+		if (!threadless_) {
+			do_timers_ = false;
+		}
 		// There's no deadline
 		return false;
 	}
@@ -335,8 +360,10 @@ bool event_loop::process_timers(scoped_lock & l)
 	auto now = monotonic_clock::now();
 	if (now < deadline_) {
 		// Deadline has not yet expired
-		do_timers_ = false;
-		timer_cond_.signal(l);
+		if (!threadless_) {
+			do_timers_ = false;
+			timer_cond_.signal(l);
+		}
 		return false;
 	}
 
@@ -394,7 +421,7 @@ bool event_loop::process_timers(scoped_lock & l)
 		return true;
 	}
 
-	if (deadline_) {
+	if (deadline_ && !threadless_) {
 		do_timers_ = false;
 		timer_cond_.signal(l);
 	}
