@@ -77,11 +77,14 @@ result file::open(native_string const& f, mode m, creation_flags d)
 	DWORD dispositionFlags;
 	if (m == writing || m == readwrite || m == appending) {
 		if (d & empty) {
-			dispositionFlags = CREATE_ALWAYS;
+			dispositionFlags = (d & nocreate) ? TRUNCATE_EXISTING : CREATE_ALWAYS;
 		}
 		else if (d & fresh) {
 			dispositionFlags = CREATE_NEW;
 		}
+		else if (d & nocreate) {
+			dispositionFlags = OPEN_EXISTING;
+		}
 		else {
 			dispositionFlags = OPEN_ALWAYS;
 		}
@@ -312,7 +315,10 @@ result file::open(native_string const& f, mode m, creation_flags d)
 			flags |= O_APPEND;
 		}
 
-		flags |= O_CREAT;
+		if (!(d & nocreate)) {
+			flags |= O_CREAT;
+		}
+
 		if (d & empty) {
 			flags |= O_TRUNC;
 		}
