@@ -35,6 +35,8 @@ size_t get_digest_size(hash_algorithm alg)
 			return 160/8;
 		case hash_algorithm::sha256:
 			return 256/8;
+		case hash_algorithm::sha384:
+			return 384/8;
 		case hash_algorithm::sha512:
 			return 512/8;
 	}
@@ -259,7 +261,7 @@ private:
 	sha256_ctx ctx_;
 };
 
-class hash_accumulator_sha512 final : public hash_accumulator::impl
+class hash_accumulator_sha512 : public hash_accumulator::impl
 {
 public:
 	hash_accumulator_sha512()
@@ -269,12 +271,12 @@ public:
 
 	virtual size_t digest_size() const override { return SHA512_DIGEST_SIZE; }
 
-	virtual void update(uint8_t const* data, size_t size) override
+	virtual void update(uint8_t const* data, size_t size) override final
 	{
 		nettle_sha512_update(&ctx_, size, data);
 	}
 
-	virtual void reinit() override
+	virtual void reinit() override final
 	{
 		nettle_sha512_init(&ctx_);
 	}
@@ -284,10 +286,22 @@ public:
 		nettle_sha512_digest(&ctx_, SHA512_DIGEST_SIZE, out);
 	}
 
-private:
+protected:
 	sha512_ctx ctx_;
 };
 
+class hash_accumulator_sha384 final : public hash_accumulator_sha512
+{
+public:
+	virtual size_t digest_size() const override { return SHA384_DIGEST_SIZE; }
+
+
+	virtual void digest(uint8_t* out) override
+	{
+		nettle_sha512_digest(&ctx_, SHA384_DIGEST_SIZE, out);
+	}
+};
+
 class hash_accumulator_hmac_sha256 final : public hash_accumulator::impl
 {
 public:
@@ -390,6 +404,9 @@ hash_accumulator::hash_accumulator(hash_algorithm algorithm)
 	case hash_algorithm::sha256:
 		impl_ = new hash_accumulator_sha256;
 		break;
+	case hash_algorithm::sha384:
+		impl_ = new hash_accumulator_sha384;
+		break;
 	case hash_algorithm::sha512:
 		impl_ = new hash_accumulator_sha512;
 		break;
