@@ -45,7 +45,9 @@ public:
 	/** \brief Creation flags when opening file for writing.
 	 *
 	 * Only evaluated when opening existing files for writing
-	 * Non-existing files will always be created when writing.
+	 * Non-existing files will always be created when writing,
+	 * except then nocreate flag is set.
+	 *
 	 * Opening for reading never creates files.
 	 */
 	enum creation_flags {
@@ -73,7 +75,13 @@ public:
 		 *
 		 * Does not modify permissions if the file already exists.
 		 */
-		 current_user_and_admins_only = 0x8
+		current_user_and_admins_only = 0x8,
+
+		/// Opening fails if the file already exists
+		fresh = 0x10,
+
+		/// Opening fails if the file does not exist
+		nocreate = 0x20
 	};
 
 	file() = default;
