@@ -69,7 +69,17 @@ public:
 	 *
 	 * The returned type can only be \c type::link if \c follow_links is \c false.
 	 */
-	static type get_file_info(native_string const& path, bool &is_link, int64_t* size, datetime* modification_time, int* mode, bool follow_links = true);
+	static type get_file_info(native_string const& path, bool& is_link, int64_t* size, datetime* modification_time, int* mode, bool follow_links = true);
+
+	/**
+	 * \brief Gets the info for the passed arguments.
+	 *
+	 * Follows symbolic links and stats the target by default, sets is_link to true if path was
+	 * a link.
+	 *
+	 * This overload additionally takes a \ref fz::local_filesys::type reference out parameter, which could be set to \c type::link only if \c follow_links is \c false.
+	 */
+	static result get_file_info(native_string const& path, bool& is_link, type& type, int64_t* size, datetime* modification_time, int* mode, bool follow_links = true);
 
 	/// Gets size of file, returns -1 on error.
 	static int64_t get_size(native_string const& path, bool *is_link = nullptr);
@@ -113,11 +123,19 @@ public:
 	void end_find_files();
 
 	static datetime get_modification_time(native_string const& path);
-	static bool set_modification_time(native_string const& path, const datetime& t);
+	static result set_modification_time(native_string const& path, const datetime& t);
 
-	/// Get the target path of a symbolic link
+	/** \brief Get the immediate target path of a symbolic link.
+	 * Only consideres the immediate link. Output may be be relative
+	 */
 	static native_string get_link_target(native_string const& path);
 
+	/** \brief Get the final target path of a symbolic link (chain)
+	 *
+	 * Result is absolute.
+	 */
+	static native_string get_final_link_target(native_string const& path);
+
 private:
 #ifdef FZ_WINDOWS
 	bool FZ_PRIVATE_SYMBOL check_buffer();
@@ -167,7 +185,7 @@ enum class mkdir_permissions
  */
 result FZ_PUBLIC_SYMBOL mkdir(native_string const& absolute_path, bool recurse, mkdir_permissions permissions = mkdir_permissions::normal, native_string * last_created = nullptr);
 
-/** \brief Removes sempty directory
+/** \brief Removes empty directory
  *
  * For recursive remove, see \ref fz::recursive_remove
  */
