@@ -6,6 +6,7 @@
 #ifdef FZ_WINDOWS
 #include "libfilezilla/glue/dll.hpp"
 #include "windows/security_descriptor_builder.hpp"
+#include <winioctl.h>
 #include <winternl.h>
 #else
 #include <errno.h>
@@ -156,6 +157,7 @@ result get_file_info_impl(result(*do_stat)(struct stat & buf, char const* path,
 		if (modification_time) {
 			*modification_time = datetime();
 		}
+		type = local_filesys::unknown;
 		return result;
 	}
 
@@ -165,7 +167,7 @@ result get_file_info_impl(result(*do_stat)(struct stat & buf, char const* path,
 
 		if (follow_links) {
 			result = do_stat(buf, path, dir, true);
-			if (result) {
+			if (!result) {
 				if (size) {
 					*size = -1;
 				}
@@ -175,6 +177,7 @@ result get_file_info_impl(result(*do_stat)(struct stat & buf, char const* path,
 				if (modification_time) {
 					*modification_time = datetime();
 				}
+				type = local_filesys::unknown;
 				return result;
 			}
 		}
@@ -287,7 +290,7 @@ result do_get_file_info(native_string const& path, bool& is_link, local_filesys:
 		if (modification_time) {
 			*modification_time = datetime();
 		}
-
+		type = local_filesys::unknown;
 		return result;
 	}
 
@@ -377,7 +380,7 @@ result do_get_file_info(native_string const& path, bool& is_link, local_filesys:
 			*size = -1;
 		}
 		type = local_filesys::dir;
-		return {result::ok,0};
+		return {result::ok, 0};
 	}
 	else {
 		if (size) {
@@ -991,7 +994,47 @@ int64_t local_filesys::get_size(native_string const& path, bool* is_link)
 	return ret;
 }
 
-native_string local_filesys::get_link_target(native_string const& path)
+#if FZ_WINDOWS
+namespace {
+bool drive_letter(wchar_t c) {
+	if (c >= 'a' && c <= 'z') {
+		return true;
+	}
+	if (c >= 'A' && c <= 'Z') {
+		return true;
+	}
+	return false;
+}
+
+// Get rid of those pesky \\?\ paths
+void sanitize_path(native_string& path)
+{
+	for (auto& c : path) {
+		if (c == '/') {
+			c = '\\';
+		}
+	}
+	while (starts_with(path, L"\\\\?")) {
+		if (starts_with(path, L"\\\\?\\UNC\\")) {
+			path = L"\\\\" + path.substr(8);
+		}
+		else if (starts_with(path, L"\\\\?\\")) {
+			if (path.size() < 6 || !drive_letter(path[4]) || path[5] != ':') {
+				path.clear();
+			}
+			else {
+				path = path.substr(4);
+			}
+		}
+		else {
+			path.clear();
+		}
+	}
+}
+}
+#endif
+
+native_string local_filesys::get_final_link_target(native_string const& path)
 {
 	native_string target;
 
@@ -1003,10 +1046,77 @@ native_string local_filesys::get_link_target(native_string const& path)
 		DWORD ret = GetFinalPathNameByHandle(hFile, out, size, 0);
 		if (ret > 0 && ret < size) {
 			target = out;
+			sanitize_path(target);
 		}
 		CloseHandle(hFile);
 	}
 #else
+	if (path.size() > 1 && path.back() == '/') {
+		return get_final_link_target(path.substr(0, path.size() - 1));
+	}
+	target.resize(PATH_MAX);
+	auto res = realpath(path.c_str(), target.data());
+	if (!res || *res != '/') {
+		return {};
+	}
+	target.resize(strlen(target.data()));
+
+#endif
+	return target;
+}
+
+native_string local_filesys::get_link_target(native_string const& path)
+{
+	native_string target;
+
+#ifdef FZ_WINDOWS
+	HANDLE hFile = CreateFile(path.c_str(), FILE_READ_EA | FILE_READ_ATTRIBUTES, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, nullptr, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT, nullptr);
+	if (hFile != INVALID_HANDLE_VALUE) {
+
+		// https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/b41f1cbf-10df-4a47-98d4-1c52a833d913
+		struct reparse_buffer {
+			ULONG tag;
+			USHORT len;
+			USHORT reserved;
+			union {
+				struct {
+					USHORT substitute_name_offset;
+					USHORT substitute_name_length;
+					USHORT print_name_offset;
+					USHORT print_name_length;
+					ULONG flags;
+					WCHAR data[1];
+				} symlink;
+				struct {
+					USHORT substitute_name_offset;
+					USHORT substitute_name_length;
+					USHORT print_name_offset;
+					USHORT print_name_length;
+					WCHAR data[1];
+				} mountpoint;
+				unsigned char dummy[2048];
+			};
+		};
+
+		reparse_buffer buf;
+		DWORD outlen{};
+		if (DeviceIoControl(hFile, FSCTL_GET_REPARSE_POINT, nullptr, 0, &buf, sizeof(buf), &outlen, nullptr) != 0) {
+			if (buf.tag == IO_REPARSE_TAG_SYMLINK) {
+				target.assign(buf.symlink.data + buf.symlink.print_name_offset / 2, buf.symlink.print_name_length / 2);
+			}
+			else if (buf.tag == IO_REPARSE_TAG_MOUNT_POINT) {
+				target.assign(buf.mountpoint.data + buf.mountpoint.print_name_offset / 2, buf.mountpoint.print_name_length / 2);
+			}
+			sanitize_path(target);
+		}
+
+		CloseHandle(hFile);
+	}
+#else
+	if (path.size() > 1 && path.back() == '/') {
+		return get_link_target(path.substr(0, path.size() - 1));
+	}
+
 	target.resize(1024);
 	while (true) {
 		ssize_t res = readlink(path.c_str(), target.data(), target.size());
