@@ -19,91 +19,148 @@
 namespace fz {
 namespace debug {
 static mutex m_;
-thread_local std::vector<std::weak_ptr<mutex_debug>> lock_stack;
+thread_local std::vector<mutex*> lock_stack;
+
+std::list<lock_order> orders;
+
 thread_local size_t waitcounter{};
 static std::ptrdiff_t mutex_offset{};
 
-void record_order(mutex* m, bool from_try)
+void FZ_PUBLIC_SYMBOL dump_orders()
 {
-	if (m == &debug::m_) {
-		return;
-	}
 	scoped_lock l(debug::m_);
 
-	auto & mdata = *m->h_.get();
-	for (auto const& wsm : lock_stack) {
-		auto sm = wsm.lock();
-		if (!sm || sm->mtx_ == m) {
-			continue;
+	std::cerr << "Known orders:\n";
+	for (auto const& order : orders) {
+		for (auto & m : order.mutexes_) {
+			std::cerr << fz::sprintf(" %p", m);
 		}
+		std::cerr << "\n";
+	}
+}
 
-		size_t i = 0;
-		while (i < sm->order_.size()) {
-			auto & o = sm->order_[i];
-			auto so = std::get<0>(o).lock();
-			if (!so) {
-				// Remove stale pointer
-				if (i + 1 != sm->order_.size()) {
-					o = std::move(sm->order_.back());
-				}
-				sm->order_.pop_back();
-				continue;
+namespace {
+bool match(lock_order const& order, std::vector<mutex*> & stack)
+{
+	if (order.mutexes_.size() != stack.size()) {
+		return false;
+	}
+	for (size_t i = 0; i < order.mutexes_.size(); ++i) {
+		if (order.mutexes_[i] != stack[i]) {
+			return false;
+		}
+	}
+
+	return true;
+}
+
+// Precondition: order contains stack.back(), its position is the pivot
+void check_inversion(lock_order const& order, std::vector<mutex*> & stack)
+{
+	size_t i{};
+	for (;; ++i) {
+		if (order.mutexes_[i] == stack.back()) {
+			break;
+		}
+		// We're still to the left of the pivot.
+
+		// Check if this a common guard mutex also on the lock stack. If that's the case, no deadlock due to inversion is possible
+		if (std::find(stack.begin(), stack.begin() + stack.size() - 1, order.mutexes_[i]) != stack.end()) {
+			return;
+		}
+	}
+
+	// We found the pivot. Check right half of order, if any of it matches, it's an inversion
+	for (size_t j = i + 1; j < order.mutexes_.size(); ++j) {
+		if (std::find(stack.begin(), stack.begin() + stack.size() - 1, order.mutexes_[j]) != stack.begin() + stack.size() - 1) {
+			std::cerr << fz::sprintf("\nLocking order violation. fz::mutex %p locked after %p\n\n", stack.back(), order.mutexes_[j]);
+
+			std::cerr << "New order:\n";
+			for (auto & m : stack) {
+				std::cerr << fz::sprintf(" %p", m);
 			}
+			std::cerr << "\n\n";
 
-			if (so->mtx_ == m) {
-				if (from_try) {
-					return;
-				}
-
+			std::cerr << "Established order:\n";
+			for (auto & m : order.mutexes_) {
+				std::cerr << fz::sprintf(" %p", m);
+			}
+			std::cerr << "\n\n";
 #if FZ_UNIX
-				std::cerr << fz::sprintf("Locking order violation. fz::mutex %p locked after %p. Reverse order was established at:\n", m, sm->mtx_);
-				auto & v = std::get<1>(o);
-				auto symbols = backtrace_symbols(v.data(), v.size());
-				if (symbols) {
-					for (size_t i = 0; i < v.size(); ++i) {
-						if (symbols[i]) {
-							std::cerr << symbols[i] << "\n";
-						}
-						else {
-							std::cerr << "unknown\n";
-						}
+			std::cerr << "Reverse order was established at:\n";
+			auto symbols = backtrace_symbols(order.backtrace_.data(), order.backtrace_.size());
+			if (symbols) {
+				for (size_t i = 0; i < order.backtrace_.size(); ++i) {
+					if (symbols[i]) {
+						std::cerr << symbols[i] << "\n";
+					}
+					else {
+						std::cerr << "unknown\n";
 					}
 				}
-				else {
-					std::cerr << "Stacktrace unavailable\n";
-				}
-#else
-				std::cerr << fz::sprintf("Locking order violation. fz::mutex %p locked after %p\n");
-#endif
-				abort();
 			}
-			++i;
+			else {
+				std::cerr << "Stacktrace unavailable\n";
+			}
+#endif
+			abort();
 		}
+	}
+}
 
-		i = 0;
-		while (i < mdata.order_.size()) {
-			auto & o = mdata.order_[i];
-			auto so = std::get<0>(o).lock();
-			if (!so) {
-				if (i + 1 != mdata.order_.size()) {
-					o = std::move(mdata.order_.back());
+void order_cleanup(mutex& m)
+{
+	scoped_lock l(debug::m_);
+	std::vector<std::list<lock_order>::iterator> own_orders;
+	std::swap(own_orders, m.debug_.own_orders_);
+	for (auto & order : own_orders) {
+		for (mutex* om : order->mutexes_) {
+			for (size_t i = 0; i < om->debug_.own_orders_.size(); ++i) {
+				if (om->debug_.own_orders_[i] == order) {
+					if (i + 1 < om->debug_.own_orders_.size()) {
+						om->debug_.own_orders_[i] = om->debug_.own_orders_.back();
+					}
+					om->debug_.own_orders_.pop_back();
+					break;
 				}
-				mdata.order_.pop_back();
-				continue;
 			}
-			if (so->mtx_ == sm->mtx_) {
-				break;
-			}
-			++i;
 		}
-		if (i == mdata.order_.size()) {
-			std::vector<void*> v;
+		orders.erase(order);
+	}
+}
+
+// Returns true if it's a new order
+void record_order(mutex& m, bool from_try)
+{
+	if (lock_stack.size() < 2) {
+		return;
+	}
+
+	scoped_lock l(debug::m_);
+	for (auto & order : m.debug_.own_orders_) {
+		if (match(*order, lock_stack)) {
+			// Order has already been seen
+			return;
+		}
+	}
+
+	// It's a new order, if not from a try_lock, check for inversion
+	if (!from_try) {
+		for (auto const& order : m.debug_.own_orders_) {
+			check_inversion(*order, lock_stack);
+		}
+	}
+
+	// Record the new order
+	orders.push_front({});
+	auto & order = orders.front();
+	order.mutexes_ = lock_stack;
 #if FZ_UNIX
-			v.resize(100);
-			v.resize(backtrace(v.data(), 100));
+	order.backtrace_.resize(100);
+	order.backtrace_.resize(backtrace(order.backtrace_.data(), 100));
 #endif
-			mdata.order_.push_back(std::make_tuple(sm, v));
-		}
+	for (auto & sm : lock_stack) {
+		sm->debug_.own_orders_.push_back(orders.begin());
 	}
 }
 
@@ -112,31 +169,37 @@ void lock(mutex* m, bool from_try) {
 		return;
 	}
 
-	if (!m->h_->count_++) {
-		lock_stack.push_back(m->h_);
-
-		if (!from_try) {
-			record_order(m, from_try);
-		}
+	if (!m->debug_.count_++) {
+		m->debug_.id_ = std::this_thread::get_id();
+		lock_stack.push_back(m);
+		record_order(*m, from_try);
 	}
 }
+}
 
 void unlock(mutex* m) {
 	if (m == &debug::m_) {
 		return;
 	}
 
-	size_t count = m->h_->count_--;
+	size_t count = m->debug_.count_--;
 	assert(count);
+	assert(m->debug_.id_ == std::this_thread::get_id());
 	if (count != 1) {
 		return;
 	}
 
-	for (auto it = lock_stack.rbegin(); it != lock_stack.rend(); ++it) {
-		auto sm = it->lock();
-		if (sm && sm->mtx_ == m) {
-			it->reset();
-			while (!lock_stack.empty() && lock_stack.back().expired()) {
+	for (size_t i = lock_stack.size() - 1; i != size_t(-1); --i) {
+		if (lock_stack[i] == m) {
+			if (i != lock_stack.size() - 1) {
+				for(; i < lock_stack.size() - 1; ++i) {
+					lock_stack[i] = lock_stack[i + 1];
+				}
+				// This may establish a new order
+				lock_stack.pop_back();
+				record_order(*m, true);
+			}
+			else {
 				lock_stack.pop_back();
 			}
 			return;
@@ -159,16 +222,18 @@ void mutex_debug::record_unlock(void* m)
 void debug_prepare_wait(void* p)
 {
 	auto m = reinterpret_cast<mutex*>(reinterpret_cast<unsigned char*>(p) - debug::mutex_offset);
-	debug::waitcounter = m->h_->count_;
+	debug::waitcounter = m->debug_.count_;
 	assert(debug::waitcounter);
-	m->h_->count_ = 0;
+	assert(m->debug_.id_ == std::this_thread::get_id());
+	m->debug_.count_ = 0;
 }
 
 void debug_post_wait(void* p)
 {
 	auto m = reinterpret_cast<mutex*>(reinterpret_cast<unsigned char*>(p) - debug::mutex_offset);
-	assert(!m->h_->count_);
-	m->h_->count_ = debug::waitcounter;
+	assert(!m->debug_.count_);
+	m->debug_.count_ = debug::waitcounter;
+	m->debug_.id_ = std::this_thread::get_id();
 }
 }
 #else
@@ -203,7 +268,7 @@ pthread_mutexattr_t* get_mutex_attributes(bool recursive)
 
 pthread_condattr_t* init_condattr()
 {
-#if HAVE_CLOCK_GETTIME && HAVE_DECL_PTHREAD_CONDATTR_SETCLOCK
+#if defined(CLOCK_MONOTONIC) && HAVE_DECL_PTHREAD_CONDATTR_SETCLOCK
 	static pthread_condattr_t attr;
 	pthread_condattr_init(&attr);
 	pthread_condattr_setclock(&attr, CLOCK_MONOTONIC);
@@ -230,14 +295,14 @@ mutex::mutex(bool recursive)
 		debug::mutex_offset = reinterpret_cast<unsigned char*>(&m_) - reinterpret_cast<unsigned char*>(this);
 		return true;
 	}();
-	h_ = std::make_shared<mutex_debug>(this);
 #endif
 }
 
 mutex::~mutex()
 {
 #ifdef LFZ_DEBUG_MUTEXES
-	assert(!h_->count_);
+	assert(!debug_.count_);
+	debug::order_cleanup(*this);
 #endif
 #ifdef FZ_WINDOWS
 	DeleteCriticalSection(&m_);
@@ -338,7 +403,7 @@ bool condition::wait(scoped_lock& l, duration const& timeout)
 #else
 
 	timespec ts;
-#if HAVE_CLOCK_GETTIME && HAVE_DECL_PTHREAD_CONDATTR_SETCLOCK
+#if defined(CLOCK_MONOTONIC) && HAVE_DECL_PTHREAD_CONDATTR_SETCLOCK
 	clock_gettime(CLOCK_MONOTONIC, &ts);
 #else
 	timeval tv{};
