@@ -1151,7 +1151,7 @@ public:
 	}
 
 	cfref(cfref const& op)
-		: ref_(op.ref)
+		: ref_(op.ref_)
 	{
 		if (ref_) {
 			CFRetain(ref_);
