@@ -22,6 +22,7 @@
   #define mutex mutex_override // Sadly on some platforms system headers include conflicting names
   #include <sys/types.h>
   #include <sys/socket.h>
+  #include <sys/un.h>
   #include <netdb.h>
   #include <fcntl.h>
   #include <unistd.h>
@@ -67,6 +68,9 @@ union sockaddr_u
 	sockaddr sockaddr_;
 	sockaddr_in in4;
 	sockaddr_in6 in6;
+#ifndef FZ_WINDOWS
+	sockaddr_un un;
+#endif
 };
 
 #if HAVE_TCP_INFO
@@ -610,7 +614,7 @@ protected:
 	// Only call while locked
 	bool do_connect(scoped_lock & l)
 	{
-		if (host_.empty() || port_.empty()) {
+		if (host_.empty()) {
 			static_cast<socket*>(socket_)->state_ = socket_state::failed;
 			return false;
 		}
@@ -622,6 +626,44 @@ protected:
 
 		sockaddr_u bindAddr{};
 
+#ifndef FZ_WINDOWS
+
+		if (socket_->family_ == AF_UNIX) {
+			host = fz::to_native(host);
+
+			sockaddr_u u{};
+			u.un.sun_family = AF_UNIX;
+			if (host.size() >= sizeof(u.un.sun_path)) {
+				if (socket_->evt_handler_) {
+					socket_->evt_handler_->send_event<socket_event>(socket_->ev_source_, socket_event_flag::connection, EINVAL);
+				}
+				static_cast<socket*>(socket_)->state_ = socket_state::failed;
+				return false;
+			}
+			strcpy(u.un.sun_path, host.c_str());
+
+			addrinfo addr{};
+			addr.ai_family = AF_UNIX;
+			addr.ai_socktype = SOCK_STREAM;
+			addr.ai_addr = &u.sockaddr_;
+			addr.ai_addrlen = sizeof(u.un);
+
+			int res = try_connect_host(addr, bindAddr, l);
+			if (res == 1) {
+				return true;
+			}
+
+			if (socket_) {
+				if (socket_->evt_handler_) {
+					socket_->evt_handler_->send_event<socket_event>(socket_->ev_source_, socket_event_flag::connection, ECONNABORTED);
+				}
+				static_cast<socket*>(socket_)->state_ = socket_state::failed;
+			}
+
+			return false;
+		}
+#endif
+
 		if (!bind.empty()) {
 			// Convert bind address
 			addrinfo bind_hints{};
@@ -1564,7 +1606,7 @@ int socket::connect(native_string const& host, unsigned int port, address_type f
 		return EISCONN;
 	}
 
-	if (port < 1 || port > 65535) {
+	if (family != address_type::unix && (port < 1 || port > 65535)) {
 		return EINVAL;
 	}
 
@@ -1585,6 +1627,11 @@ int socket::connect(native_string const& host, unsigned int port, address_type f
 	case address_type::ipv6:
 		af = AF_INET6;
 		break;
+#ifndef FZ_WINDOWS
+	case address_type::unix:
+		af = AF_UNIX;
+		break;
+#endif
 	default:
 		return EINVAL;
 	}
@@ -1594,7 +1641,7 @@ int socket::connect(native_string const& host, unsigned int port, address_type f
 
 	host_ = host;
 	port_ = port;
-	int res = socket_thread_->connect(to_utf8(host_), port_);
+	int res = socket_thread_->connect((family != address_type::unix) ? to_utf8(host_) : to_string(host_), port_);
 	if (res) {
 		state_ = socket_state::failed;
 		return res;
