@@ -1,4 +1,5 @@
 #include "libfilezilla/buffer.hpp"
+#include "libfilezilla/util.hpp"
 #include "libfilezilla/string.hpp"
 
 #ifdef FZ_WINDOWS
@@ -261,17 +262,31 @@ class iconv_t_holder
 {
 public:
 	iconv_t_holder(char const* to, char const* from)
+		: to_(to)
+		, from_(from)
 	{
-		cd = iconv_open(to, from);
+		reinit();
 	}
 
 	~iconv_t_holder()
+	{
+		close();
+	}
+
+	void close()
 	{
 		if (cd != reinterpret_cast<iconv_t>(-1)) {
 			iconv_close(cd);
+			outbuf_.wipe();
 		}
 	}
 
+	void reinit()
+	{
+		close();
+		cd = iconv_open(to_, from_);
+	}
+
 	explicit operator bool() const {
 		return cd != reinterpret_cast<iconv_t>(-1);
 	}
@@ -279,7 +294,11 @@ public:
 	iconv_t_holder(iconv_t_holder const&) = delete;
 	iconv_t_holder& operator=(iconv_t_holder const&) = delete;
 
+	char const* const to_;
+	char const* const from_;
+
 	iconv_t cd{reinterpret_cast<iconv_t>(-1)};
+	fz::buffer outbuf_;
 };
 }
 #endif
@@ -294,6 +313,33 @@ std::wstring to_wstring_from_utf8(fz::buffer const& in)
 	return to_wstring_from_utf8(reinterpret_cast<char const*>(in.get()), in.size());
 }
 
+#if FZ_WINDOWS
+void wipe_conversion_cache()
+{
+}
+#else
+namespace {
+iconv_t_holder& get_to_wide_iconv_holder()
+{
+	static thread_local iconv_t_holder holder(wchar_t_encoding(), "UTF-8");
+	return holder;
+}
+iconv_t_holder& get_to_utf8_iconv_holder()
+{
+	static thread_local iconv_t_holder holder("UTF-8", wchar_t_encoding());
+	return holder;
+}
+}
+
+void wipe_conversion_cache()
+{
+	auto & w = get_to_wide_iconv_holder();
+	w.reinit();
+	auto & u = get_to_utf8_iconv_holder();
+	u.reinit();
+}
+#endif
+
 std::wstring to_wstring_from_utf8(char const* s, size_t len)
 {
 	std::wstring ret;
@@ -308,18 +354,17 @@ std::wstring to_wstring_from_utf8(char const* s, size_t len)
 			MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, in_p, static_cast<int>(len), out_p, out_len);
 		}
 #else
-		static thread_local iconv_t_holder holder(wchar_t_encoding(), "UTF-8");
 
+		auto & holder = get_to_wide_iconv_holder();
 		if (holder && iconv(holder.cd, nullptr, nullptr, nullptr, nullptr) != static_cast<size_t>(-1)) {
 			auto in_p = const_cast<iconv_second_arg_type>(s);
 			size_t out_len = len * sizeof(wchar_t) * 2;
-			auto out_buf = std::make_unique<char[]>(out_len);
-			char* out_p = out_buf.get();
+			char* out_p = reinterpret_cast<char*>(holder.outbuf_.get(out_len));
 
 			size_t r = iconv(holder.cd, &in_p, &len, &out_p, &out_len);
 
 			if (r != static_cast<size_t>(-1)) {
-				ret.assign(reinterpret_cast<wchar_t*>(out_buf.get()), reinterpret_cast<wchar_t*>(out_p));
+				ret.assign(reinterpret_cast<wchar_t*>(holder.outbuf_.get()), reinterpret_cast<wchar_t*>(out_p));
 			}
 
 			// Our buffer should big enough as well, so we can ignore errors such as E2BIG.
@@ -409,23 +454,19 @@ std::string FZ_PUBLIC_SYMBOL to_utf8(std::wstring_view const& in)
 			WideCharToMultiByte(CP_UTF8, WC_ERR_INVALID_CHARS, in_p, static_cast<int>(in.size()), out_p, len, nullptr, nullptr);
 		}
 #else
-		static thread_local iconv_t_holder holder("UTF-8", wchar_t_encoding());
-
+		auto & holder = get_to_utf8_iconv_holder();
 		if (holder && iconv(holder.cd, nullptr, nullptr, nullptr, nullptr) != static_cast<size_t>(-1)) {
 			auto in_p = reinterpret_cast<iconv_second_arg_type>(const_cast<wchar_t*>(in.data()));
 			size_t in_len = in.size() * sizeof(wchar_t);
 
 			size_t out_len = in.size() * 4;
-			auto out_buf = std::make_unique<char[]>(out_len);
-			char* out_p = out_buf.get();
+			char* out_p = reinterpret_cast<char*>(holder.outbuf_.get(out_len));
 
 			size_t r = iconv(holder.cd, &in_p, &in_len, &out_p, &out_len);
 
 			if (r != static_cast<size_t>(-1)) {
-				ret.assign(out_buf.get(), out_p);
+				ret.assign(reinterpret_cast<char*>(holder.outbuf_.get()), out_p);
 			}
-
-			// Our buffer should big enough as well, so we can ignore errors such as E2BIG.
 		}
 #endif
 	}
