@@ -15,6 +15,29 @@
 
 #include <string.h>
 
+// In case GnuTLS has been compiled without PKCS#11 support
+#if FZ_WINDOWS
+#include "libfilezilla/glue/dll.hpp"
+
+namespace fz {
+dll& get_gnutls_dll() {
+	static dll gnutls_dll = dll::from_address(gnutls_malloc);
+	return gnutls_dll;
+}
+
+#define dynamic_gnutls_func(f) \
+	reinterpret_cast<decltype(&f)>(get_gnutls_dll()[ #f ])
+}
+
+#else
+
+#include <dlfcn.h>
+
+#define dynamic_gnutls_func(f) \
+	reinterpret_cast<decltype(&f)>(dlsym(RTLD_DEFAULT, #f))
+
+#endif
+
 using namespace std::literals;
 
 #if DEBUG_SOCKETEVENTS
@@ -162,6 +185,46 @@ public:
 		}
 	}
 	static thread_local std::function<void(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer, gnutls_x509_crl_t crl, unsigned int verification_output)> verify_output_cb_;
+
+	static int pin_callback(void *userdata, int attempt, const char *token_url, const char *token_label, unsigned int flags, char *pin, size_t pin_max)
+	{
+		tls_layer_impl::cert_context &c = *static_cast<tls_layer_impl::cert_context*>(userdata);
+
+		if (!(flags & GNUTLS_PIN_USER)) {
+			c.logger.log_u(logmsg::error, fztranslate("Wrong type of PIN requested. Flags: %d, Token: %s, URL: %s."), flags, token_label, token_url);
+			return GNUTLS_E_PKCS11_PIN_ERROR;
+		}
+
+		if (flags & GNUTLS_PIN_WRONG) {
+			c.logger.log_u(logmsg::debug_warning, fztranslate("The last tried PIN was wrong. Token: %s, URL: %s."), token_label, token_url);
+			return GNUTLS_E_PKCS11_PIN_ERROR;
+		}
+
+		if (attempt == 0) {
+			if (c.pin.size() + 1 > pin_max) {
+				c.logger.log_u(logmsg::error, fztranslate("PIN length (%d) exceeds the maximum (%d). Token: %s, URL: %s"), c.pin.size(), pin_max, token_label, token_url);
+				return GNUTLS_E_PKCS11_PIN_ERROR;
+			}
+
+			if (flags & GNUTLS_PIN_COUNT_LOW) {
+				c.logger.log_u(logmsg::debug_warning, fztranslate("Only few PIN tries remaining before the token is locked. Token: %s, URL: %s."), token_label, token_url);
+			}
+			else if (flags & GNUTLS_PIN_FINAL_TRY) {
+				c.logger.log_u(logmsg::debug_warning, fztranslate("This is the last PIN try remaining before the token is locked. Token: %s, URL: %s."), token_label, token_url);
+			}
+			else {
+				c.logger.log_u(logmsg::debug_info, fztranslate("Attempting to use PIN. Token: %s, URL: %s."), token_label, token_url);
+			}
+
+			memcpy(pin, c.pin.c_str(), c.pin.size()+1);
+
+			return 0;
+		}
+
+		c.logger.log_u(logmsg::error, fztranslate("Too many attempts (%d) while trying the PIN. Token: %s, URL: %s."), attempt+1, token_label, token_url);
+
+		return GNUTLS_E_PKCS11_PIN_ERROR;
+	}
 };
 thread_local std::function<void(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer, gnutls_x509_crl_t crl, unsigned int verification_output)> tls_layerCallbacks::verify_output_cb_;
 
@@ -262,6 +325,19 @@ tls_layer_impl::~tls_layer_impl()
 	deinit();
 }
 
+bool tls_layer_impl::cert_context::init()
+{
+	if (!credentials) {
+		int res = gnutls_certificate_allocate_credentials(&credentials);
+		if (res < 0) {
+			log_gnutls_error(res, L"gnutls_certificate_allocate_credentials"sv);
+			return false;
+		}
+	}
+
+	return true;
+}
+
 bool tls_layer_impl::init()
 {
 	// This function initializes GnuTLS
@@ -283,31 +359,35 @@ bool tls_layer_impl::init()
 #endif
 	}
 
-	if (!cert_credentials_) {
-		int res = gnutls_certificate_allocate_credentials(&cert_credentials_);
-		if (res < 0) {
-			log_error(res, L"gnutls_certificate_allocate_credentials"sv);
-			deinit();
-			return false;
-		}
+	if (!cert_context_.init()) {
+		deinit();
+		return false;
 	}
 
 	return true;
 }
 
-std::string read_certificates_file(native_string const& certsfile, logger_interface * logger)
+static std::string read_file(native_string const& path, bool cert, logger_interface * logger)
 {
-	file cf(certsfile, file::reading, file::existing);
+	file cf(path, file::reading, file::existing);
 	if (!cf.opened()) {
 		if (logger) {
-			logger->log(logmsg::error, fztranslate("Could not open certificate file."));
+			auto format = cert
+				? fztranslate("Could not open certificate file.")
+				: fztranslate("Could not open key file.");
+
+			logger->log_raw(logmsg::error, format);
 		}
 		return {};
 	}
 	int64_t const cs = cf.size();
 	if (cs < 0 || cs > 1024 * 1024) {
 		if (logger) {
-			logger->log(logmsg::error, fztranslate("Certificate file too big."));
+			auto format = cert
+				? fztranslate("Certificate file too big.")
+				: fztranslate("Key file too big.");
+
+			logger->log_raw(logmsg::error, format);
 		}
 		return {};
 	}
@@ -316,76 +396,222 @@ std::string read_certificates_file(native_string const& certsfile, logger_interf
 	auto read = cf.read2(c.data(), static_cast<size_t>(cs));
 	if (!read || read.value_ != static_cast<size_t>(cs)) {
 		if (logger) {
-			logger->log(logmsg::error, fztranslate("Could not read certificate file."));
+			auto format = cert
+				? fztranslate("Could not read certificate file.")
+				: fztranslate("Could not key file.");
+
+			logger->log_raw(logmsg::error, format);
 		}
 		return {};
 	}
 	return c;
 }
 
-bool tls_layer_impl::set_certificate_file(native_string const& keyfile, native_string const& certsfile, native_string const& password, bool pem)
+static std::string read_certificates_file(native_string const& certsfile, logger_interface * logger)
 {
-	// Load the files ourselves instead of calling gnutls_certificate_set_x509_key_file2
-	// as it takes narrow strings on MSW, thus being unable to open all files.
-
-	file kf(keyfile, file::reading, file::existing);
-	if (!kf.opened()) {
-		logger_.log(logmsg::error, fztranslate("Could not open key file."));
-		return false;
-	}
-	int64_t const ks = kf.size();
-	if (ks < 0 || ks > 1024 * 1024) {
-		logger_.log(logmsg::error, fztranslate("Key file too big."));
-		return false;
-	}
-	std::string k;
-	k.resize(ks);
-	auto read = kf.read2(k.data(), static_cast<size_t>(ks));
-	if (!read || read.value_ != static_cast<size_t>(ks)) {
-		logger_.log(logmsg::error, fztranslate("Could not read key file."));
-		return false;
-	}
-
-	std::string c = read_certificates_file(certsfile, &logger_);
-	if (c.empty()) {
-		return false;
-	}
-
-	return set_certificate(k, c, password, pem);
+	return read_file(certsfile, true, logger);
 }
 
-bool tls_layer_impl::set_certificate(std::string_view const& key, std::string_view const& certs, native_string const& password, bool pem)
+static std::string read_key_file(native_string const& keyfile, logger_interface * logger)
 {
-	if (!init()) {
+	return read_file(keyfile, false, logger);
+}
+
+static bool is_pem(std::string_view blob, tls_data_format format)
+{
+	if (format == tls_data_format::autodetect) {
+		return is_pem(blob);
+	}
+
+	return format == tls_data_format::pem;
+}
+
+unique_gnutls_privkey tls_layer_impl::import_private_key(cert_context &ctx, const_tls_param_ref key, native_string const& password, tls_data_format format)
+{
+	unique_gnutls_privkey k;
+
+	{
+		gnutls_privkey_t priv;
+		int res = gnutls_privkey_init(&priv);
+		if (res < 0) {
+			ctx.log_gnutls_error(res, L"gnutls_privkey_init"sv);
+			return nullptr;
+		}
+
+		k.reset(priv);
+	}
+
+
+	if (auto url = key.pkcs11url()) {
+		if (!*url) {
+			ctx.logger.log_raw(logmsg::error, fztranslate("Invalid pkcs11 URL in key"));
+			return nullptr;
+		}
+
+		ctx.pin = to_utf8(password);
+		gnutls_privkey_set_pin_function(k.get(), tls_layerCallbacks::pin_callback, &ctx);
+
+		int res = gnutls_privkey_import_url(k.get(), url->value.c_str(), 0);
+		if (res < 0) {
+			ctx.log_gnutls_error(res, L"gnutls_privkey_import_url"sv);
+			return nullptr;
+		}
+
+		return k;
+	}
+
+	auto import_raw = [&](unique_gnutls_privkey k, std::string_view blob) -> unique_gnutls_privkey {
+		bool pem = is_pem(blob, format);
+
+		gnutls_datum_t d;
+		d.data = const_cast<unsigned char*>(reinterpret_cast<unsigned char const*>(blob.data()));
+		d.size = blob.size();
+
+		int res = gnutls_privkey_import_x509_raw(k.get(), &d, pem ? GNUTLS_X509_FMT_PEM : GNUTLS_X509_FMT_DER,
+												 to_utf8(password).c_str(), 0);
+
+		if (res < 0) {
+			ctx.log_gnutls_error(res, L"gnutls_privkey_import_x509_raw"sv);
+			return nullptr;
+		}
+
+		return k;
+	};
+
+	if (auto keyfile = key.filepath()) {
+		if (!*keyfile) {
+			ctx.logger.log_raw(logmsg::error, fztranslate("Invalid file path for key"));
+			return nullptr;
+		}
+
+		auto blob = read_key_file(keyfile->value, &ctx.logger);
+		if (blob.empty()) {
+			return nullptr;
+		}
+
+		return import_raw(std::move(k), blob);
+	}
+
+	if (auto blob = key.blob()) {
+		if (!*blob) {
+			ctx.logger.log_raw(logmsg::error, fztranslate("Invalid data blob for key"));
+			return nullptr;
+		}
+
+		return import_raw(std::move(k), blob->value);
+	}
+
+	ctx.logger.log_raw(logmsg::error, fztranslate("Internal error: invalid key"));
+
+	return nullptr;
+}
+
+std::vector<gnutls_pcert_st> tls_layer_impl::import_certs(cert_context &ctx, const_tls_param_ref certs, tls_data_format format)
+{
+	if (certs.pkcs11url()) {
+		ctx.logger.log_raw(logmsg::error, fztranslate("PKCS11 URLs not supported yet for certificates"));
+		return {};
+	}
+
+	auto import_raw = [&](std::string_view blob) -> std::vector<gnutls_pcert_st> {
+		bool pem = is_pem(blob, format);
+
+		gnutls_datum_t d;
+		d.data = const_cast<unsigned char*>(reinterpret_cast<unsigned char const*>(blob.data()));
+		d.size = blob.size();
+
+		unsigned int num_certs = 0;
+		int res = gnutls_pcert_list_import_x509_raw(nullptr, &num_certs, &d, pem ? GNUTLS_X509_FMT_PEM : GNUTLS_X509_FMT_DER, GNUTLS_X509_CRT_LIST_IMPORT_FAIL_IF_EXCEED);
+		if (res != GNUTLS_E_SHORT_MEMORY_BUFFER) {
+			ctx.log_gnutls_error(res, L"gnutls_pcert_list_import_x509_raw count"sv);
+			return {};
+		}
+
+		std::vector<gnutls_pcert_st> pcerts(num_certs);
+		res = gnutls_pcert_list_import_x509_raw(pcerts.data(), &num_certs, &d, pem ? GNUTLS_X509_FMT_PEM : GNUTLS_X509_FMT_DER, GNUTLS_X509_CRT_LIST_IMPORT_FAIL_IF_EXCEED);
+		if (res != 0) {
+			ctx.log_gnutls_error(res, L"gnutls_pcert_list_import_x509_raw"sv);
+			return {};
+		}
+
+		return pcerts;
+	};
+
+	if (auto certsfile = certs.filepath()) {
+		if (!*certsfile) {
+			ctx.logger.log_raw(logmsg::error, fztranslate("Invalid file path for certificates"));
+			return {};
+		}
+
+		auto blob = read_certificates_file(certsfile->value, &ctx.logger);
+		if (blob.empty()) {
+			return {};
+		}
+
+		return import_raw(blob);
+	}
+
+	if (auto blob = certs.blob()) {
+		if (!*blob) {
+			ctx.logger.log_raw(logmsg::error, fztranslate("Invalid data blob for certificates"));
+			return {};
+		}
+
+		return import_raw(blob->value);
+	}
+
+	ctx.logger.log_raw(logmsg::error, fztranslate("Internal error: invalid certs"));
+
+	return {};
+}
+
+bool tls_layer_impl::set_key_and_certs(cert_context &ctx, const_tls_param_ref key, const_tls_param_ref certs, native_string const& password, tls_data_format format)
+{
+	if (!ctx.credentials) {
 		return false;
 	}
 
-	if (!cert_credentials_) {
+	auto cs = import_certs(ctx, certs, format);
+	if (cs.empty()) {
 		return false;
 	}
 
-	gnutls_datum_t c;
-	c.data = const_cast<unsigned char*>(reinterpret_cast<unsigned char const*>(certs.data()));
-	c.size = certs.size();
+	auto k = import_private_key(ctx, key, password, format);
+	if (!k) {
+		return false;
+	}
 
-	gnutls_datum_t k;
-	k.data = const_cast<unsigned char*>(reinterpret_cast<unsigned char const*>(key.data()));
-	k.size = key.size();
-
-	int res = gnutls_certificate_set_x509_key_mem2(cert_credentials_, &c,
-		&k, pem ? GNUTLS_X509_FMT_PEM : GNUTLS_X509_FMT_DER, password.empty() ? nullptr : to_utf8(password).c_str(), 0);
+	int res = gnutls_certificate_set_key(ctx.credentials, nullptr, 0, cs.data(), cs.size(), k.get());
 	if (res < 0) {
-		log_error(res, L"gnutls_certificate_set_x509_key_mem2"sv);
-		deinit();
+		ctx.log_gnutls_error(res, L"gnutls_certificate_set_key"sv);
+
+		for (auto &c: cs) {
+			gnutls_pcert_deinit(&c);
+		}
+
 		return false;
 	}
 
+	k.release();
+
+	return true;
+}
+
+bool tls_layer_impl::set_key_and_certs(const_tls_param_ref key, const_tls_param_ref certs, native_string const& password, tls_data_format format)
+{
+	if (init()) {
+		if (!set_key_and_certs(cert_context_, key, certs, password, format)) {
+			deinit();
+			return false;
+		}
+	}
+
 	return true;
 }
 
 bool tls_layer_impl::init_session(bool client, int extra_flags)
 {
-	if (!cert_credentials_) {
+	if (!cert_context_.credentials) {
 		deinit();
 		return false;
 	}
@@ -478,7 +704,7 @@ bool tls_layer_impl::init_session(bool client, int extra_flags)
 
 	gnutls_dh_set_prime_bits(session_, 1024);
 
-	gnutls_credentials_set(session_, GNUTLS_CRD_CERTIFICATE, cert_credentials_);
+	gnutls_credentials_set(session_, GNUTLS_CRD_CERTIFICATE, cert_context_.credentials);
 
 	// Setup transport functions
 	gnutls_transport_set_push_function(session_, c_push_function);
@@ -493,14 +719,19 @@ bool tls_layer_impl::init_session(bool client, int extra_flags)
 	return true;
 }
 
+void tls_layer_impl::cert_context::deinit()
+{
+	if (credentials) {
+		gnutls_certificate_free_credentials(credentials);
+		credentials = nullptr;
+	}
+}
+
 void tls_layer_impl::deinit()
 {
 	deinit_session();
 
-	if (cert_credentials_) {
-		gnutls_certificate_free_credentials(cert_credentials_);
-		cert_credentials_ = nullptr;
-	}
+	cert_context_.deinit();
 
 	if (initialized_) {
 		initialized_ = false;
@@ -1582,6 +1813,52 @@ int tls_layer_impl::get_algorithm_warnings() const
 	return algorithmWarnings;
 }
 
+int tls_layer_impl::load_certificates(const_tls_param_ref in, tls_data_format format, gnutls_x509_crt_t *& certs, unsigned int & certs_size, bool & sort, logger_interface *logger)
+{
+	if (in.pkcs11url()) {
+		if (logger) {
+			logger->log_raw(logmsg::error, fztranslate("PKCS11 URLs not supported yet for certificates"));
+		}
+
+		return {};
+	}
+
+	if (auto certsfile = in.filepath()) {
+		if (!certsfile) {
+			if (logger) {
+				logger->log_raw(logmsg::error, fztranslate("Invalid file path for certificates"));
+			}
+
+			return {};
+		}
+
+		auto blob = read_certificates_file(certsfile->value, logger);
+		if (blob.empty()) {
+			return {};
+		}
+
+		return load_certificates(blob, is_pem(blob, format), certs, certs_size, sort);
+	}
+
+	if (auto blob = in.blob()) {
+		if (!blob) {
+			if (logger) {
+				logger->log_raw(logmsg::error, fztranslate("Invalid data blob for certificates"));
+			}
+
+			return {};
+		}
+
+		return load_certificates(blob->value, is_pem(blob->value, format), certs, certs_size, sort);
+	}
+
+	if (logger) {
+		logger->log_raw(logmsg::error, fztranslate("Internal error: invalid certs"));
+	}
+
+	return {};
+}
+
 int tls_layer_impl::load_certificates(std::string_view const& in, bool pem, gnutls_x509_crt_t *& certs, unsigned int & certs_size, bool & sort)
 {
 	gnutls_datum_t dpem;
@@ -1805,7 +2082,7 @@ int tls_layer_impl::verify_certificate()
 			gnutls_credentials_set(session_, GNUTLS_CRD_CERTIFICATE, cred);
 			unsigned int status = 0;
 			int verifyResult = gnutls_certificate_verify_peers3(session_, to_utf8(hostname_).c_str(), &status);
-			gnutls_credentials_set(session_, GNUTLS_CRD_CERTIFICATE, cert_credentials_);
+			gnutls_credentials_set(session_, GNUTLS_CRD_CERTIFICATE, cert_context_.credentials);
 			std::get<1>(lease).unlock();
 
 			gnutls_session_set_verify_output_function(session_, nullptr);
@@ -1869,7 +2146,7 @@ int tls_layer_impl::verify_certificate()
 			}
 
 			gnutls_x509_trust_list_t tlist;
-			gnutls_certificate_get_trust_list(cert_credentials_, &tlist);
+			gnutls_certificate_get_trust_list(cert_context_.credentials, &tlist);
 			if (gnutls_x509_trust_list_add_cas(tlist, &root, 1, 0) != 1) {
 				logger_.log(logmsg::error, fztranslate("Could not add certificate to temporary trust list"));
 				failure(0, true);
@@ -1878,7 +2155,7 @@ int tls_layer_impl::verify_certificate()
 
 			// 2. Also disable time checks. We allow expired/not yet valid certificates, though only
 			// after explicit user confirmation.
-			gnutls_certificate_set_verify_flags(cert_credentials_, gnutls_certificate_get_verify_flags(cert_credentials_) | GNUTLS_VERIFY_DISABLE_TIME_CHECKS | GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS);
+			gnutls_certificate_set_verify_flags(cert_context_.credentials, gnutls_certificate_get_verify_flags(cert_context_.credentials) | GNUTLS_VERIFY_DISABLE_TIME_CHECKS | GNUTLS_VERIFY_DISABLE_TRUSTED_TIME_CHECKS);
 
 			unsigned int status = 0;
 			int verifyResult = gnutls_certificate_verify_peers2(session_, &status);
@@ -2286,6 +2563,11 @@ bool verify_purposes(std::set<std::string> & purposes, tls_layer::cert_type t)
 				return {};
 			}
 		}
+		else if (p == GNUTLS_KP_CODE_SIGNING) {
+			if (t != tls_layer::cert_type::any && t != tls_layer::cert_type::codesign) {
+				return {};
+			}
+		}
 		else {
 			return {};
 		}
@@ -2294,10 +2576,49 @@ bool verify_purposes(std::set<std::string> & purposes, tls_layer::cert_type t)
 	return true;
 }
 
-bool set_cert_params(gnutls_x509_crt_t & crt, tls_layer::cert_type t, std::string const& distinguished_name, std::vector<std::string> const& hostnames, duration lifetime)
+bool set_purposes_from_type(tls_layer_impl::cert_context &ctx, gnutls_x509_crt_t & crt, tls_layer::cert_type t)
+{
+	std::set<std::string> purposes;
+	if (!get_purposes(crt, purposes)) {
+		ctx.logger.log_raw(logmsg::error, fztranslate("Could not get certificate purposes."));
+		return {};
+	}
+	if (!verify_purposes(purposes, t)) {
+		ctx.logger.log_raw(logmsg::error, fztranslate("Certificate purposes do not match with its intended type."));
+		return {};
+	}
+	char const* oid{};
+	switch (t) {
+	case tls_layer::cert_type::client:
+		oid = GNUTLS_KP_TLS_WWW_CLIENT;
+		break;
+	case tls_layer::cert_type::server:
+		oid = GNUTLS_KP_TLS_WWW_SERVER;
+		break;
+	case tls_layer::cert_type::codesign:
+		oid = GNUTLS_KP_CODE_SIGNING;
+		break;
+	default:
+		break;
+	}
+	if (oid) {
+		if (purposes.find(oid) == purposes.end()) {
+			int res = gnutls_x509_crt_set_key_purpose_oid(crt, oid, 1);
+			if (res) {
+				ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_key_purpose_oid");
+				return {};
+			}
+		}
+	}
+
+	return true;
+}
+
+static bool set_cert_params(tls_layer_impl::cert_context &ctx, gnutls_x509_crt_t & crt, tls_layer::cert_type t, std::string const& distinguished_name, std::vector<std::string> const& hostnames, duration lifetime)
 {
 	int res = gnutls_x509_crt_set_serial(crt, random_bytes(20).data(), 20);
 	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_serial");
 		return {};
 	}
 
@@ -2305,6 +2626,7 @@ bool set_cert_params(gnutls_x509_crt_t & crt, tls_layer::cert_type t, std::strin
 		char const* out{};
 		int res = gnutls_x509_crt_set_dn(crt, distinguished_name.c_str(), &out);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_dn");
 			return {};
 		}
 	}
@@ -2313,6 +2635,7 @@ bool set_cert_params(gnutls_x509_crt_t & crt, tls_layer::cert_type t, std::strin
 		for (size_t i = 0; i < hostnames.size(); ++i) {
 			res = gnutls_x509_crt_set_subject_alt_name(crt, GNUTLS_SAN_DNSNAME, hostnames[i].c_str(), hostnames[i].size(), i ? GNUTLS_FSAN_APPEND : GNUTLS_FSAN_SET);
 			if (res) {
+				ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_subject_alt_name");
 				return {};
 			}
 		}
@@ -2326,10 +2649,12 @@ bool set_cert_params(gnutls_x509_crt_t & crt, tls_layer::cert_type t, std::strin
 
 	res = gnutls_x509_crt_set_activation_time(crt, (now - duration::from_minutes(5)).get_time_t());
 	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_activation_time");
 		return {};
 	}
 	res = gnutls_x509_crt_set_expiration_time(crt, (now + lifetime).get_time_t());
 	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_expiration_time");
 		return {};
 	}
 
@@ -2337,57 +2662,112 @@ bool set_cert_params(gnutls_x509_crt_t & crt, tls_layer::cert_type t, std::strin
 	if (t == tls_layer::cert_type::ca) {
 		usage = GNUTLS_KEY_KEY_CERT_SIGN | GNUTLS_KEY_CRL_SIGN;
 	}
+	else if (t == tls_layer::cert_type::codesign) {
+		usage = GNUTLS_KEY_DIGITAL_SIGNATURE;
+	}
 	else {
 		usage = GNUTLS_KEY_DIGITAL_SIGNATURE | GNUTLS_KEY_KEY_ENCIPHERMENT;
 	}
 
 	res = gnutls_x509_crt_set_key_usage(crt, usage);
 	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_key_usage");
 		return {};
 	}
 
 
 	res = gnutls_x509_crt_set_basic_constraints(crt, (t == tls_layer::cert_type::ca) ? 1 : 0, -1);
 	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_basic_constraints");
 		return {};
 	}
 
-	std::set<std::string> purposes;
-	if (!get_purposes(crt, purposes)) {
+	if (!set_purposes_from_type(ctx, crt, t)) {
 		return {};
 	}
-	if (!verify_purposes(purposes, t)) {
-		return false;
-	}
-	if (t == tls_layer::cert_type::client || t == tls_layer::cert_type::server) {
-		auto p = (t == tls_layer::cert_type::server) ? GNUTLS_KP_TLS_WWW_SERVER : GNUTLS_KP_TLS_WWW_CLIENT;
-		if (purposes.find(p) == purposes.end()) {
-			res = gnutls_x509_crt_set_key_purpose_oid(crt, p, 1);
-			if (res) {
-				return {};
-			}
-		}
-	}
+
 	return true;
 }
 }
 
-std::pair<std::string, std::string> tls_layer_impl::generate_selfsigned_certificate(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, duration const& lifetime, tls_layer::cert_type type, bool ecdsa)
+std::string tls_layer_impl::generate_selfsigned_certificate(cert_context &ctx, unique_gnutls_privkey priv, std::string const& distinguished_name, std::vector<std::string> const& hostnames, duration const& lifetime, tls_layer::cert_type type)
+{
+	gnutls_x509_crt_t crt;
+	int res = gnutls_x509_crt_init(&crt);
+	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crt_init");
+		return {};
+	}
+
+	auto ret = [&]() -> std::string {
+		res = gnutls_x509_crt_set_version(crt, 3);
+		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_version");
+			return {};
+		}
+
+		{
+			gnutls_pubkey_t pub;
+			res = gnutls_pubkey_init(&pub);
+			if (res) {
+				ctx.log_gnutls_error(res, L"gnutls_pubkey_init");
+			}
+			else {
+				unsigned int key_usage = (type == tls_layer::cert_type::ca) ? (GNUTLS_KEY_KEY_CERT_SIGN | GNUTLS_KEY_CRL_SIGN) : (GNUTLS_KEY_DIGITAL_SIGNATURE | GNUTLS_KEY_KEY_ENCIPHERMENT);
+
+				res = gnutls_pubkey_import_privkey(pub, priv.get(), key_usage, 0);
+				if (res) {
+					ctx.log_gnutls_error(res, L"gnutls_pubkey_import_privkey");
+				}
+				else {
+					res = gnutls_x509_crt_set_pubkey(crt, pub);
+					if (res) {
+						ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_pubkey");
+					}
+				}
+				gnutls_pubkey_deinit(pub);
+			}
+		}
+
+		if (res) {
+			return {};
+		}
+
+		if (!set_cert_params(ctx, crt, type, distinguished_name, hostnames, lifetime)) {
+			return {};
+		}
+
+		res = gnutls_x509_crt_privkey_sign(crt, crt, priv.get(), GNUTLS_DIG_SHA256, 0);
+		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_privkey_sign");
+			return {};
+		}
+
+		datum_holder ch;
+		res = gnutls_x509_crt_export2(crt, GNUTLS_X509_FMT_PEM, &ch);
+		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_export2");
+			return {};
+		}
+
+		return ch.to_string();
+	}();
+
+	gnutls_x509_crt_deinit(crt);
+
+	return ret;
+}
+
+static std::pair<unique_gnutls_privkey, std::string> generate_priv_key_for_cert(tls_layer_impl::cert_context &ctx, native_string const& password, tls_layer::cert_type type, bool ecdsa)
 {
 	gnutls_x509_privkey_t priv;
 	int res = gnutls_x509_privkey_init(&priv);
 	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_privkey_init");
 		return {};
 	}
 
-	gnutls_x509_crt_t crt;
-	res = gnutls_x509_crt_init(&crt);
-	if (res) {
-		gnutls_x509_privkey_deinit(priv);
-		return {};
-	}
-
-	auto ret = [&]() -> std::pair<std::string, std::string> {
+	auto ret = [&]() -> std::pair<unique_gnutls_privkey, std::string> {
 
 		auto fmt = ecdsa ? GNUTLS_PK_ECDSA : GNUTLS_PK_RSA;
 		unsigned int bits = gnutls_sec_param_to_pk_bits(fmt, GNUTLS_SEC_PARAM_HIGH);
@@ -2400,6 +2780,7 @@ std::pair<std::string, std::string> tls_layer_impl::generate_selfsigned_certific
 
 		res = gnutls_x509_privkey_generate(priv, fmt, bits, 0);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_privkey_generate");
 			return {};
 		}
 
@@ -2407,68 +2788,108 @@ std::pair<std::string, std::string> tls_layer_impl::generate_selfsigned_certific
 
 		if (password.empty()) {
 			res = gnutls_x509_privkey_export2(priv, GNUTLS_X509_FMT_PEM, &kh);
+			if (res) {
+				ctx.log_gnutls_error(res, L"gnutls_x509_privkey_export2");
+				return {};
+			}
 		}
 		else {
 			res = gnutls_x509_privkey_export2_pkcs8(priv, GNUTLS_X509_FMT_PEM, to_utf8(password).c_str(), 0, &kh);
+			if (res) {
+				ctx.log_gnutls_error(res, L"gnutls_x509_privkey_export2_pkcs8");
+				return {};
+			}
 		}
+
+		gnutls_privkey_t abstract_priv;
+		res = gnutls_privkey_init(&abstract_priv);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_privkey_init");
 			return {};
 		}
 
-		res = gnutls_x509_crt_set_version(crt, 3);
+		res = gnutls_privkey_import_x509(abstract_priv, priv, GNUTLS_PRIVKEY_IMPORT_AUTO_RELEASE);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_privkey_import_x509");
+			gnutls_privkey_deinit(abstract_priv);
 			return {};
 		}
 
-		res = gnutls_x509_crt_set_key(crt, priv);
-		if (res) {
-			return {};
-		}
+		priv = nullptr;
 
-		if (!set_cert_params(crt, type, distinguished_name, hostnames, lifetime)) {
-			return {};
-		}
-
-		res = gnutls_x509_crt_sign2(crt, crt, priv, GNUTLS_DIG_SHA256, 0);
-		if (res) {
-			return {};
-		}
-
-		datum_holder ch;
-		res = gnutls_x509_crt_export2(crt, GNUTLS_X509_FMT_PEM, &ch);
-		if (res) {
-			return {};
-		}
-
-		return {kh.to_string(), ch.to_string()};
+		return {unique_gnutls_privkey(abstract_priv), kh.to_string()};
 	}();
 
-	gnutls_x509_privkey_deinit(priv);
-	gnutls_x509_crt_deinit(crt);
+	if (priv) {
+		gnutls_x509_privkey_deinit(priv);
+	}
 
 	return ret;
 }
 
-std::string tls_layer_impl::generate_cert_from_csr(std::pair<std::string, std::string> const& issuer, native_string const& password, std::string const& csr, std::string const& distinguished_name, std::vector<std::string> const& hostnames, duration const& lifetime, tls_layer::tls_layer::cert_type type)
+std::pair<std::string, std::string> tls_layer_impl::generate_selfsigned_certificate(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, duration const& lifetime, tls_layer::cert_type type, bool ecdsa, logger_interface &logger)
 {
+	cert_context ctx(logger);
+
+	auto [priv, key] = generate_priv_key_for_cert(ctx, password, type, ecdsa);
+	if (!priv) {
+		return {};
+	}
+
+	auto cert = generate_selfsigned_certificate(ctx, std::move(priv), distinguished_name, hostnames, lifetime, type);
+	if (cert.empty()) {
+		return {};
+	}
+
+	return {std::move(key), std::move(cert)};
+}
+
+std::string tls_layer_impl::generate_selfsigned_certificate(const_tls_param_ref key, native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, duration const& lifetime, tls_layer::cert_type type, logger_interface &logger)
+{
+	cert_context ctx(logger);
+
+	auto priv = import_private_key(ctx, key, password, tls_data_format::autodetect);
+	if (!priv) {
+		return {};
+	}
+
+	auto cert = generate_selfsigned_certificate(ctx, std::move(priv), distinguished_name, hostnames, lifetime, type);
+	if (cert.empty()) {
+		return {};
+	}
+
+	return cert;
+}
+
+std::string tls_layer_impl::generate_cert_from_csr(std::pair<std::string, std::string> const& issuer, native_string const& password, std::string const& csr, std::string const& distinguished_name, std::vector<std::string> const& hostnames, duration const& lifetime, tls_layer::tls_layer::cert_type type, logger_interface &logger)
+{
+	cert_context ctx(logger);
+
 	gnutls_x509_privkey_t priv;
 	int res = gnutls_x509_privkey_init(&priv);
 	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_privkey_init");
 		return {};
 	}
 	gnutls_x509_crt_t icrt;
-	if (gnutls_x509_crt_init(&icrt)) {
+	res = gnutls_x509_crt_init(&icrt);
+	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crt_init");
 		gnutls_x509_privkey_deinit(priv);
 		return {};
 	}
 	gnutls_x509_crq_t crq;
-	if (gnutls_x509_crq_init(&crq) != 0) {
+	res = gnutls_x509_crq_init(&crq);
+	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crq_init");
 		gnutls_x509_crt_deinit(icrt);
 		gnutls_x509_privkey_deinit(priv);
 		return {};
 	}
 	gnutls_x509_crt_t crt;
-	if (gnutls_x509_crt_init(&crt) != 0) {
+	res = gnutls_x509_crt_init(&crt);
+	if (res) {
+		ctx.log_gnutls_error(res, L"gnutls_x509_crt_init");
 		gnutls_x509_crq_deinit(crq);
 		gnutls_x509_crt_deinit(icrt);
 		gnutls_x509_privkey_deinit(priv);
@@ -2480,7 +2901,9 @@ std::string tls_layer_impl::generate_cert_from_csr(std::pair<std::string, std::s
 		gnutls_datum_t kd;
 		kd.data = const_cast<unsigned char*>(reinterpret_cast<unsigned char const*>(issuer.first.data()));
 		kd.size = issuer.first.size();
-		if (gnutls_x509_privkey_import2(priv, &kd, GNUTLS_X509_FMT_PEM, to_utf8(password).c_str(), 0) != 0) {
+		res = gnutls_x509_privkey_import2(priv, &kd, GNUTLS_X509_FMT_PEM, to_utf8(password).c_str(), 0);
+		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_privkey_import2");
 			return {};
 		}
 
@@ -2489,6 +2912,7 @@ std::string tls_layer_impl::generate_cert_from_csr(std::pair<std::string, std::s
 		icrtd.size = issuer.second.size();
 		res = gnutls_x509_crt_import(icrt, &icrtd, GNUTLS_X509_FMT_PEM);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_import");
 			return {};
 		}
 
@@ -2497,20 +2921,24 @@ std::string tls_layer_impl::generate_cert_from_csr(std::pair<std::string, std::s
 		csrd.size = csr.size();
 		res = gnutls_x509_crq_import(crq, &csrd, GNUTLS_X509_FMT_PEM);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_import");
 			return {};
 		}
 
 		res = gnutls_x509_crq_verify(crq, 0);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_verify");
 			return {};
 		}
 
 		// Verify key purpose oids
 		std::set<std::string> purposes;
 		if (!get_purposes(crq, purposes)) {
+			ctx.logger.log_raw(logmsg::error, fztranslate("Could not get certificate request purposes."));
 			return {};
 		}
 		if (!verify_purposes(purposes, type)) {
+			ctx.logger.log_raw(logmsg::error, fztranslate("Certificate request purposes do not match with its type."));
 			return {};
 		}
 
@@ -2518,47 +2946,55 @@ std::string tls_layer_impl::generate_cert_from_csr(std::pair<std::string, std::s
 		unsigned int usage{};
 		unsigned int critical{};
 		if ((res = gnutls_x509_crq_get_key_usage(crq, &usage, &critical)) && res != GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_get_key_usage");
 			return {};
 		}
 
 		if (type == tls_layer::cert_type::ca) {
 			if (usage & ~(GNUTLS_KEY_KEY_CERT_SIGN | GNUTLS_KEY_CRL_SIGN)) {
+				ctx.logger.log_raw(logmsg::error, fztranslate("Certificate request has wrong key usage types set."));
 				return {};
 			}
 		}
 		else {
 			if (usage & ~(GNUTLS_KEY_DIGITAL_SIGNATURE | GNUTLS_KEY_KEY_ENCIPHERMENT)) {
+				ctx.logger.log_raw(logmsg::error, fztranslate("Certificate request has wrong key usage types set."));
 				return {};
 			}
 		}
 
 		res = gnutls_x509_crt_set_version(crt, 3);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_version");
 			return {};
 		}
 
 		res = gnutls_x509_crt_set_crq(crt, crq);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_crq");
 			return {};
 		}
 
 		res = gnutls_x509_crt_set_crq_extensions(crt, crq);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_set_crq_extensions");
 			return {};
 		}
 
-		if (!set_cert_params(crt, type, distinguished_name, hostnames, lifetime)) {
+		if (!set_cert_params(ctx, crt, type, distinguished_name, hostnames, lifetime)) {
 			return {};
 		}
 
 		res = gnutls_x509_crt_sign(crt, icrt, priv);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_sign");
 			return {};
 		}
 
 		datum_holder ch;
 		res = gnutls_x509_crt_export2(crt, GNUTLS_X509_FMT_PEM, &ch);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crt_export2");
 			return {};
 		}
 
@@ -2573,51 +3009,46 @@ std::string tls_layer_impl::generate_cert_from_csr(std::pair<std::string, std::s
 	return cert;
 }
 
-std::pair<std::string, std::string> tls_layer_impl::generate_csr(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, bool csr_as_pem, tls_layer::cert_type type)
+std::string tls_layer_impl::generate_csr(cert_context &ctx, unique_gnutls_privkey priv, std::string const& distinguished_name, std::vector<std::string> const& hostnames, bool csr_as_pem, tls_layer::cert_type type)
 {
-	gnutls_x509_privkey_t priv;
-	int res = gnutls_x509_privkey_init(&priv);
-	if (res) {
-		return {};
-	}
 	gnutls_x509_crq_t crq;
-	res = gnutls_x509_crq_init(&crq);
+	int res = gnutls_x509_crq_init(&crq);
 	if (res) {
-		gnutls_x509_privkey_deinit(priv);
+		ctx.log_gnutls_error(res, L"gnutls_x509_crq_init");
 		return {};
 	}
 
-	auto ret = [&]() -> std::pair<std::string, std::string> {
-
-		auto fmt = GNUTLS_PK_ECDSA;
-		unsigned int bits = gnutls_sec_param_to_pk_bits(fmt, GNUTLS_SEC_PARAM_HIGH);
-		if (fmt == GNUTLS_PK_RSA && bits < 2048) {
-			bits = 2048;
-		}
-
-		res = gnutls_x509_privkey_generate(priv, fmt, bits, 0);
-		if (res) {
-			return {};
-		}
-
-		datum_holder kh;
-
-		if (password.empty()) {
-			res = gnutls_x509_privkey_export2(priv, GNUTLS_X509_FMT_PEM, &kh);
-		}
-		else {
-			res = gnutls_x509_privkey_export2_pkcs8(priv, GNUTLS_X509_FMT_PEM, to_utf8(password).c_str(), 0, &kh);
-		}
-		if (res) {
-			return {};
-		}
-
+	auto ret = [&]() -> std::string {
 		res = gnutls_x509_crq_set_version(crq, 3);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_set_version");
 			return {};
 		}
 
-		res = gnutls_x509_crq_set_key(crq, priv);
+		unsigned int key_usage = (type == tls_layer::cert_type::ca) ? (GNUTLS_KEY_KEY_CERT_SIGN | GNUTLS_KEY_CRL_SIGN) : (GNUTLS_KEY_DIGITAL_SIGNATURE | GNUTLS_KEY_KEY_ENCIPHERMENT);
+
+		{
+			gnutls_pubkey_t pub;
+			res = gnutls_pubkey_init(&pub);
+			if (res) {
+				ctx.log_gnutls_error(res, L"gnutls_pubkey_init");
+			}
+			else {
+				res = gnutls_pubkey_import_privkey(pub, priv.get(), key_usage, 0);
+				if (res) {
+					ctx.log_gnutls_error(res, L"gnutls_pubkey_import_privkey");
+				}
+				else {
+					res = gnutls_x509_crq_set_pubkey(crq, pub);
+					if (res) {
+						ctx.log_gnutls_error(res, L"gnutls_x509_crq_set_pubkey");
+					}
+				}
+				gnutls_pubkey_deinit(pub);
+			}
+		}
+
+
 		if (res) {
 			return {};
 		}
@@ -2625,53 +3056,119 @@ std::pair<std::string, std::string> tls_layer_impl::generate_csr(native_string c
 		char const* out{};
 		res = gnutls_x509_crq_set_dn(crq, distinguished_name.c_str(), &out);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_set_dn");
 			return {};
 		}
 
 		for (auto const& hostname : hostnames) {
 			res = gnutls_x509_crq_set_subject_alt_name(crq, GNUTLS_SAN_DNSNAME, hostname.c_str(), hostname.size(), GNUTLS_FSAN_APPEND);
 			if (res) {
+				ctx.log_gnutls_error(res, L"gnutls_x509_crq_set_subject_alt_name");
 				return {};
 			}
 		}
 
-		res = gnutls_x509_crq_set_key_usage(crq, (type == tls_layer::cert_type::ca) ? (GNUTLS_KEY_KEY_CERT_SIGN | GNUTLS_KEY_CRL_SIGN) : (GNUTLS_KEY_DIGITAL_SIGNATURE | GNUTLS_KEY_KEY_ENCIPHERMENT));
+		res = gnutls_x509_crq_set_key_usage(crq, key_usage);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_set_key_usage");
 			return {};
 		}
 
 		if (type == tls_layer::cert_type::client || type == tls_layer::cert_type::server) {
 			res = gnutls_x509_crq_set_key_purpose_oid(crq, (type == tls_layer::cert_type::server) ? GNUTLS_KP_TLS_WWW_SERVER : GNUTLS_KP_TLS_WWW_CLIENT, 1);
 			if (res) {
+				ctx.log_gnutls_error(res, L"gnutls_x509_crq_set_key_purpose_oid");
 				return {};
 			}
 		}
 
 		res = gnutls_x509_crq_set_basic_constraints(crq, 0, -1);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_set_basic_constraints");
 			return {};
 		}
 
-		res = gnutls_x509_crq_sign2(crq, priv, GNUTLS_DIG_SHA256, 0);
+		res = gnutls_x509_crq_privkey_sign(crq, priv.get(), GNUTLS_DIG_SHA256, 0);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_privkey_sign");
 			return {};
 		}
 
 		datum_holder ch;
 		res = gnutls_x509_crq_export2(crq, csr_as_pem ? GNUTLS_X509_FMT_PEM : GNUTLS_X509_FMT_DER, &ch);
 		if (res) {
+			ctx.log_gnutls_error(res, L"gnutls_x509_crq_export2");
 			return {};
 		}
 
-		return {kh.to_string(), ch.to_string()};
+		return ch.to_string();
 	}();
 
 	gnutls_x509_crq_deinit(crq);
-	gnutls_x509_privkey_deinit(priv);
 
 	return ret;
 }
 
+std::string tls_layer_impl::generate_csr(const_tls_param_ref key, native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, bool csr_as_pem, tls_layer::cert_type type, logger_interface &logger)
+{
+	cert_context ctx(logger);
+
+	auto priv = import_private_key(ctx, key, password, tls_data_format::autodetect);
+	if (!priv) {
+		return {};
+	}
+
+	return generate_csr(ctx, std::move(priv), distinguished_name, hostnames, csr_as_pem, type);
+}
+
+std::pair<std::string, std::string> tls_layer_impl::generate_csr(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, bool csr_as_pem, tls_layer::cert_type type, bool ecdsa, logger_interface &logger)
+{
+	cert_context ctx(logger);
+
+	auto [priv, key] = generate_priv_key_for_cert(ctx, password, type, ecdsa);
+	if (!priv) {
+		return {};
+	}
+
+	auto cert = generate_csr(ctx, std::move(priv), distinguished_name, hostnames, csr_as_pem, type);
+	if (cert.empty()) {
+		return {};
+	}
+
+	return {std::move(key), std::move(cert)};
+}
+
+bool tls_layer_impl::add_pkcs11_provider(const native_string_view &path, logger_interface &logger)
+{
+	static auto const pgnutls_pkcs11_init = dynamic_gnutls_func(gnutls_pkcs11_init);
+	static auto const pgnutls_pkcs11_add_provider = dynamic_gnutls_func(gnutls_pkcs11_add_provider);
+
+	if (!pgnutls_pkcs11_init || !pgnutls_pkcs11_add_provider) {
+		logger.log(logmsg::error, fztranslate("GnuTLS has been compiled without PKCS#11 support"));
+		return {};
+	}
+
+	int res = pgnutls_pkcs11_init(GNUTLS_PKCS11_FLAG_MANUAL, nullptr);
+	if (res) {
+		log_gnutls_error(logger, res, L"gnutls_pkcs11_init");
+		return {};
+	}
+
+	res = pgnutls_pkcs11_add_provider(to_string(path).c_str(), nullptr);
+	if (res) {
+		log_gnutls_error(logger, res, L"gnutls_pkcs11_add_provider");
+		return {};
+	}
+
+	return true;
+}
+
+bool tls_layer::supports_pkcs11()
+{
+	static bool const supported = dynamic_gnutls_func(gnutls_pkcs11_init) != nullptr;
+	return supported;
+}
+
 int tls_layer_impl::shutdown_read()
 {
 	if (!can_read_from_socket_) {
