@@ -6,6 +6,7 @@
 #ifdef FZ_WINDOWS
 
 #include "../libfilezilla/glue/windows.hpp"
+#include "../libfilezilla/logger.hpp"
 #include <memory>
 
 namespace fz {
@@ -55,8 +56,10 @@ private:
 	std::unique_ptr<impl> impl_;
 };
 
-std::string GetSidFromToken(HANDLE h);
+bool GetUserAndDomainFromSid(PSID sid, std::wstring& user, std::wstring& domain);
+bool GetUserInfoFromToken(HANDLE h, std::string& sid_string, std::wstring& roaming_profile_path, std::wstring& username, std::wstring& domain, logger_interface& logger = get_null_logger());
 bool DropAdminPrivilegesFromToken(HANDLE h);
+
 }
 
 #endif
